const net = require('net');
const WebSocket = require('ws');

// Datos del MasterServer
const masterserverIp = 'localhost'; // Cambiar por la IP del masterserver
const masterserverPort = 3000; // Puerto del masterserver

// Datos del servidor de chat
const serverIp = '0.0.0.0'; // Cambiar por la IP Publica(externa) del servidor de chat.
const listenPort = 5555; // Puerto del servidor de chat para escuchar conexiones.
const sendListenPort = 5555; // Puerto publico(externo) del servidor de chat(Cambiarlo si es diferente).
const wsPort = 5556; // Puerto para el WebSocket
const serverName = 'Mi servidor de chat'; // Nombre del servidor de chat.
let useMasterServer = false; // Puedes cambiar esto a false para desactivar la conexión con el Master Server

// Array para almacenar los clientes conectados
let clients = [];
let masterConnected = false;
let reconnectInterval;

// Función para conectar al MasterServer
function connectToMaster() {
    if (useMasterServer && !masterConnected) {
        console.log('Intentando conectar al MasterServer...');
        masterClient.connect({ port: masterserverPort, host: masterserverIp });
    }
}

const emptyness = ''

// Crear un cliente TCP para conectarse al MasterServer
const masterClient = new net.Socket();

// Manejar la conexión con el MasterServer
masterClient.on('connect', () => {
    console.log('Conexión establecida con el MasterServer');
    masterConnected = true;
    clearInterval(reconnectInterval);
    // Enviar la información al MasterServer
    masterClient.write(JSON.stringify({ ip: serverIp, port: sendListenPort, name: serverName }));
});

// Manejar los datos recibidos del MasterServer
masterClient.on('data', (data) => {
    console.log('Respuesta del MasterServer:', data.toString());
});

// Manejar la desconexión con el MasterServer
masterClient.on('close', () => {
    console.log('Conexión con el MasterServer cerrada');
    masterConnected = false;
    // Intentar reconexión cada minuto si useMasterServer es verdadero
    if (useMasterServer) {
        reconnectInterval = setInterval(connectToMaster, 60000);
    }
});

// Manejar el error en la conexión con el MasterServer
masterClient.on('error', (error) => {
    console.error('Error en la conexión con el MasterServer:', error.message);
    masterConnected = false;
    // Intentar reconexión cada minuto si useMasterServer es verdadero
    if (useMasterServer) {
        reconnectInterval = setInterval(connectToMaster, 60000);
    }
});

// Crear un servidor TCP
const server = net.createServer(socket => {
    // Manejar los datos recibidos del cliente
    socket.on('data', data => {
        const message = data.toString().trim();
        if (!socket.userName && message) {
            socket.userName = message; // Asignar el nombre de usuario al socket
            console.log(`Usuario conectado: ${socket.userName}`);
        } else {
            broadcast(`${message}`, socket);
        }
        // Actualizar el tiempo del último mensaje recibido y el nombre de usuario en el array clients
        const clientIndex = clients.findIndex(client => client.socket === socket);
        if (clientIndex !== -1) {
            clients[clientIndex].lastSeen = Date.now();
            clients[clientIndex].userName = socket.userName;
        }
    });

    // Manejar el evento 'close' del cliente
    socket.on('close', () => {
        const disconnectedUser = clients.find(client => client.socket === socket);
        if (disconnectedUser) {
            clients = clients.filter(client => client.socket !== socket); // Eliminar el cliente desconectado del array
            console.log(`Usuario desconectado: ${disconnectedUser.userName}`);
            broadcastUserList();
        }
    });

    // Manejar el evento 'error' del cliente
    socket.on('error', error => {
        console.error(`Error en el cliente ${socket.remoteAddress}:${socket.remotePort}: ${error.message}`);
    });

    clients.push({ socket, userName: null, lastSeen: Date.now() }); // Agregar el nuevo cliente al array con la marca de tiempo actual
    broadcastUserList();
});

// Crear un servidor WebSocket
const wss = new WebSocket.Server({ port: wsPort });

wss.on('connection', function (ws) {
    ws.on('message', function (message) {
        message = message.toString().trim();
        const data = JSON.parse(message);
        if (data.type === 'connect') {
            ws.userName = data.userName;
            console.log(`Usuario conectado via WebSocket: ${ws.userName}`);
            broadcastUserList();
        } else if (data.type === 'message') {
            broadcast(`${ws.userName}: ${data.message}`, ws);
        }
    });

    ws.on('close', function () {
        const disconnectedUser = clients.find(client => client.socket === ws);
        if (disconnectedUser) {
            clients = clients.filter(client => client.socket !== ws); // Eliminar el cliente desconectado del array
            console.log(`Usuario desconectado via WebSocket: ${disconnectedUser.userName}`);
            broadcastUserList();
        }
    });

    ws.on('error', function (error) {
        console.error(`Error en el cliente WebSocket: ${error.message}`);
    });

    clients.push({ socket: ws, userName: null, lastSeen: Date.now() });
    broadcastUserList();
});

// Función para difundir un mensaje a todos los clientes excepto al remitente
function broadcast(message, sender) {
    clients.forEach(client => {
        if (client.socket !== sender && !client.socket.destroyed) {
            client.socket.send ? client.socket.send(JSON.stringify({ type: 'message', userName: sender.userName, message })) : client.socket.write(message + '\n');
        }
    });
    process.stdout.write(message + '\n'); // Imprimir el mensaje en la consola del servidor con un salto de línea
}

// Función para enviar la lista de usuarios a todos los clientes
function broadcastUserList() {
    const userNames = clients
        .filter(client => client.userName)
        .map(client => client.userName);
    const userPackage = JSON.stringify({ type: 'users', users: userNames });
    clients.forEach(client => {
        if (!client.socket.destroyed) {
            client.socket.send ? client.socket.send(userPackage) : client.socket.write(userPackage + '\n');
        }
    });
}

// Enviar el paquete de usuarios cada 5 segundos
setInterval(broadcastUserList, 5000);

// Función para verificar y eliminar clientes inactivos
setInterval(() => {
    const now = Date.now();
    clients = clients.filter(client => {
        if (now - client.lastSeen > 600000) { // 10 minutos de inactividad
            console.log(`Desconectando usuario inactivo: ${client.userName}`);
            client.socket.end ? client.socket.end() : client.socket.close(); // Cerrar la conexión con el cliente inactivo
            return false;
        }
        return true;
    });
}, 5000); // Verificar cada 5 segundos

// Manejar el evento 'error' del servidor
server.on('error', error => {
    console.error(`Error en el servidor: ${error.message}`);
});

// Escuchar en el puerto listenPort
server.listen(listenPort, () => {
    console.log(`Servidor de chat escuchando en el puerto ${listenPort}`);
});

console.log(`Servidor WebSocket funcionando en puerto 5556`)

// Intentar conectar al MasterServer al inicio si useMasterServer es verdadero
if (useMasterServer) {
    connectToMaster();
}
